import argparse
import sys
import textwrap

from . import Archive

__version__ = 'wbsv 0.2.5'


def natural_num(n):
    """Judge whether numstr is positive or not."""
    if int(n) < 1:
        print("[!]Err: num {} should be more than 0.".format(n),
              file=sys.stderr)
        exit(1)
    return n


def parse_args():
    """Parse arguments."""
    parser = argparse.ArgumentParser(
        prog='wbsv',
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=textwrap.dedent('''\
            CLI tool for save webpage on Wayback Machine forever.
            Save webpage and one's all URI(s) on Wayback Machine.'''),
        epilog=textwrap.dedent('''\
            additional information:
                If you don't give the URL,
                interactive mode will be launched.
                (To quit interactive mode,
                 type "end", "exit", "exit()",
                 "break", "bye", ":q" or "finish".)'''))

    parser.add_argument('url',  metavar='url', nargs='*',
                        help='Saving pages in order.')
    parser.add_argument('-V', '--version', action='version',
                        version=__version__,
                        help='Show version and exit')
    parser.add_argument('-r', '--retry', default=3,
                        metavar='cnt', type=int,
                        help='Set a retry limit on failed save.')
    parser.add_argument('-t', '--only_target', action='store_true',
                        help='Save just target webpage(s).')
    parser.add_argument('-L', '--level', default=1,
                        metavar='lv', type=int,
                        help='Set maximum recursion depth.')
    parser.add_argument('-d', '--dry_run', action='store_true', default=False,
                        help='Running without saving (dry-run mode)')
    args = parser.parse_args()
    urls = [i for i in args.url if Archive.is_url(i)]
    if args.url != urls:
        print("[!]invalid url format", file=sys.stderr)
        exit(1)

    param = {
        "retry": natural_num(args.retry),
        "urls": urls,
        "only-target": args.only_target,
        "dry-run": args.dry_run,
        "errout": sys.stderr,
        "out": sys.stdout,
        "level": natural_num(args.level)}
    return param



from urllib.request import urlopen
from urllib.error import HTTPError, URLError
from http.client import IncompleteRead
import sys

from bs4 import BeautifulSoup

from urllib.parse import urlparse, urljoin


def is_page(url):
    """Judge whether str is webpage."""
    exclude_suffixes = (".css", ".gif", ".jpeg", ".jpg",
                        ".js", ".json", ".png", ".svg")
    url_parts = urlparse(url)
    return not url_parts.fragment and not url_parts.path.endswith(
        exclude_suffixes)


def is_valid_scheme(url):
    """Judge whether url is valid scheme."""
    return urlparse(url).scheme in ["ftp", "gopher", "http", "https"]


def remove_useless(l):
    """Remove not available data from list."""
    return {x for x in l if x is not None and len(x) > 1}


def find_uri(url):
    """Find links in page."""
    # extract elements containing of uri links in a page
    try:
        html_source = urlopen(url)

    except (HTTPError, URLError, UnicodeEncodeError) as e:
        print("[!]" + str(type(e)), file=sys.stderr)
        return set()

    html_source_charset = html_source.headers.get_content_charset(
        failobj="utf-8")

    try:
        html_decoded = html_source.read().decode(
            html_source_charset, 'ignore')
        uris_misc = BeautifulSoup(html_decoded, "html.parser").findAll(
            ["a"])

    except IncompleteRead:
        uris_misc = []

    # extract uri link data
    uris_misc = sum([[i.get("href")] for i in uris_misc], [])
    # change "relative" uri into "absolute" one
    uris_misc = {urljoin(url, i) for i in uris_misc}
    # exclude mailto:// or javascript:// ...
    uris_misc = {i for i in uris_misc if is_valid_scheme(i) and is_page(i)}
    return remove_useless(uris_misc)


def extract_uri_recursive(url, rec):
    """Extract uri links from a page."""
    uri_dic = {url}
    search_queue = [[url]]
    if rec == 0:
        uri_dic = find_uri(url)

    for lev in range(rec):
        print("[+]LEVEL: %d" % (lev + 1))
        add_dic = set()
        remain_count = 1
        for url in search_queue[-1]:
            print("[+]REMAIN:%d/%d" %
                  (remain_count, len(search_queue[-1])), end="\r")
            add_dic |= find_uri(url)
            remain_count += 1

        print("[+]LEVEL: %d FINISHED!" % (lev + 1))
        uri_dic |= add_dic
        search_queue.append(add_dic)

    return uri_dic



import re
import sys
import time
import random

from requests.exceptions import ConnectionError, TooManyRedirects, HTTPError

from savepagenow import capture_or_cache
from savepagenow.api import WaybackRuntimeError


def is_url(url):
    """Judge whether str is url or not."""
    return re.compile(r'^(http|https)://').match(url)


def is_end(url):
    """Judge whether str is the fin cmd to quit interactive mode."""
    return re.compile(r'^(end|exit|exit\(\)|break|bye|:q|finish)$').match(url)


def show_err():
    """Print error texts without stopping process in error."""
    for err in list(sys.exc_info()):
        err_msg = "[!]%s" % str(err).strip("<>")
        print(err_msg, file=sys.stderr)


def wait_min():
    """Wait a minute."""
    for t in range(60):
        print("%d/60s" % t, end="\r", file=sys.stderr)
        time.sleep(1)


def retry_ntime(ret, func, c, uri):
    """Retry function given times."""
    for cnt in range(ret):
        if func:
            return True
        else:
            print("[%s/%d]: Retrying..." % (c[0], c[1]),
                  "COUNT:%d" % (cnt + 1), end="\r")
    print("[%s/%d]:" % (c[0], c[1]), "<FAIL> %s" % uri)
    return False


def add_res(func, t, f):
    """Increment t or f by func bool."""
    return [t+1, f] if func else [t, f+1]


def try_archive(id_, dic_size, uri):
    """Try to save a page on Wayback Machine."""
    try:
        print("[%s/%d]: Wait...    " % (id_, dic_size), end="\r")
        time.sleep(random.uniform(1, 3))
        archive_uri, exist_f = capture_or_cache(
            uri,
            user_agent='Mozilla/5.0 (iPhone; CPU iPhone OS 10_2 like Mac OS X) '\
                       'AppleWebKit/602.3.12 (KHTML, like Gecko) Version/10.0 '\
                       'Mobile/14C92 Safari/602.1'
        )
        print("[%s/%d]:" % (id_, dic_size), end=" ")
        print("<%s>" % "NOW" if exist_f else "PAST", archive_uri)
        return True

    except WaybackRuntimeError:
        return False


def archive(uri_dic, pageurl, RETRY, DRY):
    """Save URIs extracted from the target page."""
    print("[+]Now: %s" % pageurl)
    print("[+]%d URI(s) found." % len(uri_dic))
    if DRY:
        [print(uri) for uri in uri_dic]
        return True
    # try to throw each uri to API
    count, saves, fails = 0, 0, 0
    for uri in uri_dic:
        count += 1
        # count uri_dic RETRY
        id_ = str(count).zfill(len(str(len(uri_dic))))

        try:
            saves, fails = add_res(
                retry_ntime(RETRY,
                            try_archive(id_, len(uri_dic), uri),
                            (id_, len(uri_dic)),
                            uri),
                saves, fails)

        except KeyboardInterrupt:
            print("[!]Interrupted!", file=sys.stderr)
            print("[!]Halt.", file=sys.stderr)
            break

        except (ConnectionError, TooManyRedirects, HTTPError) as e:
            print("[!]API says: " + str(type(e)), file=sys.stderr)
            print("[!]Need a 1 min break...", file=sys.stderr)
            wait_min()

    # after for-loop
    print("[+]FIN!: %s" % pageurl)
    print("[+]ALL:", count, "SAVE:", saves, "FAIL:", fails)



import sys

from . import Archive
from . import Find


def get_input(v):
    """Get a REPL input."""
    try:
        if Archive.is_url(v):
            return v

        else:
            print("[!]This input is invalid.", file=sys.stderr)
            return ''

    except(EOFError, KeyboardInterrupt):
        print("\n[+]End.")
        exit(0)


def chk_url_cond(url, opt):
    """Change a behavior according to options."""
    if Archive.is_end(url):
        print("[+]End.")
        exit(0)

    elif opt["only-target"]:
        Archive.archive([url], url, opt["retry"], opt["dry-run"])

    else:
        try:
            dic = Find.extract_uri_recursive(url, opt["level"])

        except KeyboardInterrupt:
            dic = []
            print("[!]Interrupted!", file=sys.stderr)
            print("[!]Halt.", file=sys.stderr)

        Archive.archive(dic, url, opt["retry"], opt["dry-run"])


def interactive(opt):
    """Interactive mode like shell."""
    print("[+]To exit, use CTRL+C or type 'end'")
    while True:
        print("[[Input a target url (ex: https://google.com)]]")
        url = get_input(input(">>> "))
        if not url:
            chk_url_cond(url, opt)



import sys

from . import Archive
from . import Find
from . import ParseArgs
from . import Interact


def iter_urls(opt):
    """Iterate given urls for saving."""
    try:
        for x in opt["urls"]:
            Archive.archive(Find.extract_uri_recursive(x, opt["level"]),
                            x, opt["retry"], opt["dry-run"])

    except KeyboardInterrupt:
        print("[!]Interrupted!", file=sys.stderr)
        print("[!]Halt.", file=sys.stderr)
        exit(1)


def main():
    """Main function."""
    opt = ParseArgs.parse_args()

    if len(opt["urls"]) == 0:
        Interact.interactive(opt)

    elif opt["only-target"]:
        [Archive.archive([x], x, opt["retry"], opt["dry-run"]) for x in opt["urls"]]
        exit(0)

    else:
        iter_urls(opt)
        exit(0)


if __name__ == "__main__":
    main()



import sys
import re

class Interactive:
    def __init__(self, finder, archiver):
        """initialize."""
        self.finder = finder
        self.archiver = archiver

    def parse_opt(self, opt):
        self.only_target = opt["only-target"]

    @staticmethod
    def is_end(str):
        """Judge whether str is the fin cmd to quit interactive mode."""
        return re.compile(r'^(end|exit|exit\(\)|break|bye|:q|finish)$').match(str)

    @staticmethod
    def is_url(url):
        """Judge whether str is url or not."""
        return re.compile(r'\A(http|https)://').match(url)

    def judge_str(self, str):
        if self.is_end(str):
            print("[+]End.")
            exit(0)
        elif self.is_url(str):
            self.finder.find_and_archive(str, self.archiver)
            self.finder.print_result()
        else:
            print("[!]This input is invalid.", file=sys.stderr)

    def run(self):
        """Interactive mode like shell."""
        print("[+]To exit, use CTRL+C or type 'end'")

        while True:
            print("[[Input a target url (ex: https://google.com)]]")

            try:
                str = input(">>> ")
                self.judge_str(str)
            except(EOFError, KeyboardInterrupt):
                print("\n[+]End.")
                exit(0)



import sys

from .archiver import RandomArchiver, Archiver
from .finder import Finder
from . import ParseArgs
from .interactive import Interactive


def collect_archiver(opt):
    if opt["dry-run"] :
        return RandomArchiver()
    else:
        return Archiver()

def main():
    """Main function."""
    opt = ParseArgs.parse_args()

    archiver = collect_archiver(opt)
    finder = Finder()

    archiver.parse_opt(opt)
    finder.parse_opt(opt)

    if len(opt["urls"]) == 0:
        interact = Interactive(finder, archiver)
        interact.parse_opt(opt)
        interact.run()

    else:
        try:
            [finder.find_and_archive(x, archiver) for x in opt["urls"]]
            archiver.print_result()
        except KeyboardInterrupt:
            print("[!]Interrupted!", file=sys.stderr)
            print("[!]Halt.", file=sys.stderr)
            exit(1)


if __name__ == "__main__":
    main()



import time

from random import random, randint

from savepagenow import capture_or_cache
from savepagenow.api import WaybackRuntimeError


class AbstractArchiver():
    def __init__(self):
        """initialize."""
        self.save_count = 0
        self.fail_count = 0

    def parse_opt(self, opt):
        self.retry = opt["retry"]

    def try_archive(self, url):
        pass

    def archive(self, url):
        print("Saving: "+url)
        result = self.retry_ntimes( self.try_archive, [url] )
        self.add_result(result)
        print( ("Saved: " if result else "Failed: ")+url)

    def retry_ntimes(self, func, args=()):
        for i in range(self.retry):
            if func(*args):
                return True
            else:
                print("fail: "+str(i+1))
                time.sleep(randint(2,10))

        return False

    def add_result(self, result):
        if result:
            self.save_count += 1
        else:
            self.fail_count += 1

    def print_result(self):
        #print("[+]FIN!: %s" % pageurl)
        print("SAVE:", self.save_count, "FAIL:", self.fail_count)

class Archiver(AbstractArchiver):
    def try_archive(self, url):
        try:
            #print("[%s/%d]: Wait...    " % (id_, dic_size), end="\r")
            archive_uri, exist_f = capture_or_cache(url)
            #print("[%s/%d]:" % (id_, dic_size), end=" ")
            print("<%s>" % "NOW" if exist_f else "PAST", archive_uri)
            return True

        except WaybackRuntimeError:
            return False

class RandomArchiver(AbstractArchiver):
    def try_archive(self, url):
        if random() > 0.5 :
            return True
        else:
            return False



from urllib.request import urlopen
from urllib.error import HTTPError, URLError
from http.client import IncompleteRead
from random import randint
import time

from bs4 import BeautifulSoup

from urllib.parse import urlparse, urljoin


class Finder:
    def __init__(self):
        """initialize."""
        self.urls = set()
        self.fetched_urls = set()

    def clean(self):
        self.urls = set()

    def print_result(self):
        print("Fetched:", len(self.fetched_urls), "URLs, ",
            "Found:", len(self.urls), "URLs")

    def parse_opt(self, opt):
        self.search_url_depth = opt["level"]
        if opt["only-target"] :
            self.search_url_depth = 0

    @staticmethod
    def is_page(url):
        """Judge whether str is webpage."""
        exclude_suffixes = (".css", ".gif", ".jpeg", ".jpg",
                            ".js", ".json", ".png", ".svg")
        url_parts = urlparse(url)
        return not url_parts.fragment and not url_parts.path.endswith(
            exclude_suffixes)

    @staticmethod
    def is_valid_scheme(url):
        """Judge whether url is valid scheme."""
        return urlparse(url).scheme in ["ftp", "gopher", "http", "https"]

    @staticmethod
    def remove_useless(l):
        """Remove not available data from list."""
        return {x for x in l if x is not None and len(x) > 1}

    def find_uri(self, url):
        """Find links in page."""
        # extract elements containing of uri links in a page
        if url in self.fetched_urls :
            return set()
        else:
            self.fetched_urls.add(url)

        try:
            print("Fetching: "+url)
            html_source = urlopen(url)
        except (HTTPError, URLError, UnicodeEncodeError):
            return set()

        html_source_charset = html_source.headers.get_content_charset(
            failobj="utf-8")

        try:
            html_decoded = html_source.read().decode(
                html_source_charset, 'ignore'
            )
            uris_misc = BeautifulSoup(html_decoded, "html.parser").findAll(
                ["a"]
            )
        except IncompleteRead:
            uris_misc = []

        # extract uri link data
        uris_misc = sum([[i.get("href")] for i in uris_misc], [])
        # change "relative" uri into "absolute" one
        uris_misc = {urljoin(url, i) for i in uris_misc}
        # exclude mailto:// or javascript:// ...
        uris_misc = {i for i in uris_misc if self.is_valid_scheme(i) and self.is_page(i)}
        return self.remove_useless(uris_misc)

    def extract_uri_recursive(self, url, depth):
        # to exit extracting URL, Listen KeyboardInterrupt arround this method.
        if depth==0 :
            self.urls.add(url)
            return
        elif depth==1 :
            self.urls |= self.find_uri(url)
            return
        else:
            included_urls = self.find_uri(url)
            for l in included_urls :
                self.extract_uri_recursive(l, depth-1)
                time.sleep(randint(2,5))

            self.urls |= included_urls

    def find_and_archive(self, url, archiver):
        try:
            self.extract_uri_recursive(url, self.search_url_depth)
        except KeyboardInterrupt:
            print("Interrupt: Stopped extracting url.")

        self.print_result()

        try:
            for l in self.urls :
                archiver.archive(l)
        except KeyboardInterrupt:
            print("Interrupt: Stopped saving pages.")



from setuptools import setup, find_packages
setup(
    name="wbsv",
    version="0.2.5",
    description="Throw all URIs in a page on to Wayback Machine from CLI.",
    description_content_type="",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url="https://github.com/eggplants/wbsv-cli",
    author="eggplants",
    packages=find_packages(),
    include_package_data=True,
    license='MIT',
    install_requires=["beautifulsoup4>=4.9", "savepagenow>=1.0"],
    entry_points={
        "console_scripts": [
            "wbsv=wbsv.main:main"
        ]
    }
)
